/*
 * epspd.c version 1.3 part of the vfloppy 1.4 package
 *
 * Copyright 1996 Justin Mitchell (madmitch@discordia.org.uk) and friends.
 *
 * This version delivered in 2002 by Fred Jan Kraan (fjkraan@xs4all.nl)
 *
 * vfread is placed under the GNU General Public License in July 2002.
 *
 *  This file is part of Vfloppy 1.4.
 *
 *  Vfloppy is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Vfloppy is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Vfloppy; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/*
 * epspd - emulates up to four virtual disks for the Epson PX-4 and PX-8
 *         computers using the epsp protocol. 
 * epspd -s <port> [-0 <img0>] [-1 <img1>] [-2 <img2>] [-3 <img3>] [-d <debuglevel>]
 */

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include <termios.h>
#include <unistd.h>
#include <fcntl.h>
#include "epspv3.h"
#include "fdc.h"
#include "epspHeaders.h"

int epsp_port;
int drive_fd[4];
int debug = 0;
/*int unit0 = 0;
int unit1 = 0; */
int units[2];
int commandForMe = 0;
int dataSize = 0;

struct driveParameters driveParam;

void initDpData() {
	int i;
	for (i=0; i < BLOCK_SIZE; i++)
		driveParam.data[i] = 0;
}
void initLabel() {
	char *empty = "";
	int i;
	for (i=0; i < LABEL_SIZE; i++) {
		label[i] = empty;
	}
	label[STANDARD_WRITE]   = "Standard write";	/* 0x00 */
	label[FLUSH_WRITE]      = "Flush write";	/* 0x01 */
	label[SEQUENTIAL_WRITE] = "Sequential write";	/* 0x02 */
	label[SID_MAPLE]    = "PX-8 (22h)";		/* 0x22 */
	label[SID_PINE]     = "PX-4 (23h)";		/* 0x23 */
	label[DID_DE]       = "unit D/E (31h)";		/* 0x31 */
	label[DID_FG]       = "unit F/G (32h)";		/* 0x32 */
	label[FDC_RESET]    = "PINE_RESET (0Dh)";	/* 0x0D */
	label[FDC_RESET_M]  = "MAPLE_RESET (0Eh)";	/* 0X0E */
	label[FDC_READ]     = "FDC_READ (77h)";		/* 0x77 */
	label[FDC_WRITE]    = "FDC_WRITE (78h)";	/* 0x78 */
	label[FDC_WRITEHST] = "FDC_WRITEHST (79h)";	/* 0x79 */
	label[FDC_COPY]     = "FDC_COPY (7Ah)";		/* 0x7A */
	label[FDC_FORMAT]   = "FDC_FORMAT (7Ch)";	/* 0x7C */
}

void initDeviceNames() {
	char *empty = "";
	int i;
	for (i = 0; i < DEVICE_NAME_SIZE; i++) {
		deviceNames[i] = empty;
	}
}

void setSerial(char *s_device) {
	struct termios t; 
        if ( s_device[0] != '\0' ) {
                printf("\nSerial port %s specified\n", s_device);
                epsp_port = open(s_device, O_RDWR);
                if(epsp_port == -1)
                {
			printf(" setSerial; Error opening serial port %s. Exiting...\n", s_device);
                        perror(s_device);
                        exit(2);
                } else {
                        if (debug > 2) 
                        	printf(" setSerial; Serial port \'%s\' (fd: %d) opened succesfully.\n", 
					s_device, epsp_port);
			deviceNames[epsp_port] = s_device;
                }
        } else {
                printf("No serial port specified. Exiting...\n");
                exit(1);
        }

        if(tcgetattr(epsp_port, &t)==-1) {
                perror("tcgetattr");
                printf(" setSerial; Some error with the serial port\n");
                exit(2);
        }

        cfsetispeed(&t, B38400);
        cfsetospeed(&t, B38400);
        t.c_iflag=t.c_iflag&~(ISTRIP|INLCR|ICRNL|IGNCR|IUCLC|IXON|IXANY|IXOFF);
        t.c_oflag=t.c_oflag&~(OPOST);
        t.c_cflag=t.c_cflag&~(CSIZE|PARENB);
        t.c_cflag|=CS8;
        t.c_lflag=t.c_lflag&~(ISIG|ICANON|ECHO);
        t.c_cc[VTIME]=0;
        t.c_cc[VMIN]=1;

        if(tcsetattr(epsp_port, TCSANOW, &t)==-1) {
                perror("tcsetattr");
                exit(2);
        }
}

void mountDiskImage(char *dImage, int driveId, char *unitName) {
	int fd = open(dImage, O_RDWR);
	if (fd == -1) {
		printf("Some error with opening %s. Exiting...\n", dImage);
		perror(dImage);
		exit(2);
	}
	int diskNo = driveId & 1;	/* Hack */
	drive_fd[driveId] = fd;
	deviceNames[fd] = dImage;
	printf("Unit: %s, Disk: %d, image file: \'%s\', (fd: %d) mounted ok.\n", unitName, diskNo, dImage, fd);
}

void session() {
	while(1) {
		if (waitForEOT()) {
			if (debug > 2) printf(" session; EOT received at start cycle, expecting address message\n");
		} else {
			printf("WARN: received char not EOT, restarting\n");
			continue;
		} 
		if (recENQBlock()) {
			if (debug > 2) printf(" session; ENQ block ok, sending ACK, expecting command message\n");
			sendACK();
		} else {
			printf("WARN: ENQ block not ok, sending NAK\n");
			sendNAK();
			continue;
		}
		if (recHeaderBlock()) {
			if (debug > 2) printf(" session; Header block ok, sending ACK, expecting data message\n");
			sendACK();
		} else {
			printf("WARN: Header block not ok, sending NAK\n");
			sendNAK();
			continue;
		}	
		if (recTextBlock()) {
			if (debug > 2) printf(" session; Text block ok, sending ACK, expecting EOT\n");
			sendACK();
		} else {
			printf("WARN: error in Text block size, sending NAK\n");
			sendNAK();
			continue;
		}
		if (waitForEOT()) {
			if (debug > 2) printf(" session; EOT received, to send mode\n");
		} else {
			printf("WARN: received char not EOT, restarting\n");
			continue;
		} 
		sendHeaderBlock();
		if (debug > 2) printf(" session; Header Block sent\n");
		if (waitForACK()) {
			if (debug > 2) printf(" session; ACK received, sending Text Block\n");
		} else {
			continue;
		}
		sendTextBlock();
		if (debug > 2) printf(" session; Text Block sent\n");
		if (waitForACK()) {
			if (debug > 2) printf(" session; ACK received, session completed, sending EOT\n");
		} else {
			continue;
		}
		sendEOT();
		if (debug > 2) printf(" session; EOT sent, waiting for EOT\n");
		if (waitForEOT()) 
			if (debug > 2) printf(" session; EOT received, at end cycle\n");
		if (debug > 2) printf("--------------------------------------------------------------------------------\n"); 
	}	
}

int main(int argc, char *argv[]) {
	char ch;
	char *optstr = "d:0:1:2:3:s:";

	initLabel();
	initDeviceNames();
	initDpData();
	units[0] = 0;
	units[1] = 0;
	drive_fd[0] = -1;
	drive_fd[1] = -1;
	drive_fd[2] = -1;
	drive_fd[3] = -1;

	if (debug > 1) {
		printf("\nEPSPD version 1.5 (2009-0m-dd)");
		printf("\nDebug level: %d", debug);
	}

	while( -1 != (ch=getopt(argc,argv,optstr))) {
		switch(ch) {
			case '0':           
			/*	printf("option: -0 : %s\n", optarg); */
				units[0] = 1;
				mountDiskImage(optarg, 0, label[DID_DE]);
				break;
			case '1':           
			/*	printf("option: -1 : %s\n", optarg); */
				units[0] = 1;
				mountDiskImage(optarg, 1, label[DID_DE]);
				break;
			case '2':           
			/*	printf("option: -2 : %s\n", optarg); */
				units[1] = 1;
				mountDiskImage(optarg, 2, label[DID_FG]);
				break;
			case '3':           
			/*	printf("option: -3 : %s\n", optarg); */
				units[1] = 1;
				mountDiskImage(optarg, 3, label[DID_FG]);
				break;
			case 's':           
			/*	printf("option: -s : %s\n", optarg); */
				setSerial(optarg);
				break;
			case 'd':
				debug = atoi(optarg);
			/*	printf("option: -d : %d\n", debug); */
				printf(" main; set debug level to %d\n", debug);
				break;
			case '?':
				printf(" main; unrecognized option: %c\n",optopt);
				break;
			default:
				printf(" main; error?  condition unaccounted for?\n");
				break;

		}
	}
	
	/*
	 *	Up.. time to go; 'Clunk, waitForIt..., 'Clunk', 'Clunk'
	 */
	 
	printf("\nEPSPD: Starting EPSP disk services on %s", deviceNames[epsp_port]);
	printf("\n");

	session();
	return(1);
}
